import slacker
from jsonview.exceptions import BadRequest
from rest_framework import generics
from rest_framework import permissions

from orchestra.core.errors import TaskAssignmentError
from orchestra.core.errors import WorkerCertificationError
from orchestra.interface_api.project_management.decorators import \
    is_project_admin
from orchestra.interface_api.project_management.decorators import \
    project_management_api_view
from orchestra.interface_api.project_management import project_management
from orchestra.models import Task
from orchestra.models import Project
from orchestra.models import Worker
from orchestra.utils.load_json import load_encoded_json
from orchestra.project_api.serializers import ProjectSerializer
from orchestra.utils.revert import revert_task_to_iteration
from orchestra.utils.task_lifecycle import complete_and_skip_task
from orchestra.utils.task_lifecycle import create_subsequent_tasks
from orchestra.utils.task_lifecycle import reassign_assignment
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import end_project


class IsProjectAdmin(permissions.BasePermission):
    def has_permission(self, request, view):
        return is_project_admin(request.user)


class ProjectList(generics.ListCreateAPIView):
    permission_classes = (permissions.IsAuthenticated, IsProjectAdmin)
    serializer_class = ProjectSerializer
    queryset = Project.objects.exclude(status=Project.Status.ABORTED)


@project_management_api_view
def project_information_api(request):
    project_id = load_encoded_json(request.body)['project_id']
    try:
        return project_management.project_management_information(project_id)
    except Project.DoesNotExist:
        raise BadRequest('Project not found for the given id.')


@project_management_api_view
def reassign_assignment_api(request):
    worker_username = load_encoded_json(request.body)['worker_username']
    try:
        worker = Worker.objects.get(user__username=worker_username)
    except Worker.DoesNotExist:
        raise BadRequest('Worker not found for the given username.')
    assignment_id = load_encoded_json(request.body)['assignment_id']

    try:
        reassign_assignment(worker.id, assignment_id)
    except (WorkerCertificationError, TaskAssignmentError) as e:
        raise BadRequest(e)


@project_management_api_view
def revert_task_api(request):
    body = load_encoded_json(request.body)
    try:
        audit = revert_task_to_iteration(
            body['task_id'], body['iteration_id'],
            body.get('revert_before'), body.get('commit'))
    except Task.DoesNotExist as e:
        raise BadRequest(e)
    return audit


@project_management_api_view
def complete_and_skip_task_api(request):
    task_id = load_encoded_json(request.body)['task_id']
    complete_and_skip_task(task_id)


@project_management_api_view
def assign_task_api(request):
    worker_username = load_encoded_json(request.body)['worker_username']
    try:
        worker = Worker.objects.get(user__username=worker_username)
        task_id = load_encoded_json(request.body)['task_id']
        assign_task(worker.id, task_id)
    except (Worker.DoesNotExist,
            Task.DoesNotExist,
            WorkerCertificationError) as e:
        raise BadRequest(e)


@project_management_api_view
def create_subsequent_tasks_api(request):
    project_id = load_encoded_json(request.body)['project_id']
    try:
        project = Project.objects.get(id=project_id)
    except Project.DoesNotExist:
        raise BadRequest('Project not found for the given id.')
    create_subsequent_tasks(project)


@project_management_api_view
def edit_slack_membership_api(request):
    body = load_encoded_json(request.body)
    try:
        project_management.edit_slack_membership(
            body['project_id'], body['username'], body['action'])
    except slacker.Error as e:
        raise BadRequest(e)


@project_management_api_view
def end_project_api(request):
    project_id = load_encoded_json(request.body)['project_id']
    try:
        end_project(project_id)
    except Project.DoesNotExist:
        raise BadRequest('Project not found for the given id.')



from io import StringIO
import os
import sys

from django.core.management import call_command
from django.test import modify_settings

from orchestra.models import Workflow
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.workflow import assert_test_dir_v1_loaded
from orchestra.tests.helpers.workflow import assert_test_dir_workflow_loaded
from orchestra.workflow.load import load_workflow


NONEXISTENT_WORKFLOW_SLUG = 'DOESNOTEXIST'
TEST_WORKFLOWS_MODULE = 'orchestra.tests.workflows'
TEST_WORKFLOW = 'test_dir'
TEST_WORKFLOW_DIR = os.path.abspath(os.path.join(
    os.path.dirname(__file__),
    '../workflows/{}'.format(TEST_WORKFLOW)))

VERSION_1 = 'test_v1'
NONEXISTENT_VERSION = 'DOESNOTEXIST'


@modify_settings(INSTALLED_APPS={
    'append': '{}.{}'.format(TEST_WORKFLOWS_MODULE, TEST_WORKFLOW)
})
class LoadWorkflowSampleDataTestCase(OrchestraTestCase):

    def test_loadworkflowsampledata(self):
        """ Ensure that workflow sample data loads correctly. """

        # Load the test workflow
        load_workflow(TEST_WORKFLOW, VERSION_1)
        assert_test_dir_workflow_loaded(self)
        assert_test_dir_v1_loaded(self)

        # Calling the command with an invalid version argument should fail.
        invalid_args = (
            VERSION_1,  # Workflow slug omitted
            TEST_WORKFLOW,  # Version slug omitted
            '{}/{}/'.format(TEST_WORKFLOW, VERSION_1),  # Too many slashes
            '{}.{}'.format(TEST_WORKFLOW, VERSION_1),  # Wrong separator
        )
        for invalid_arg in invalid_args:
            invalid_stderr = StringIO()
            invalid_message = 'Please specify workflow versions in the format'
            call_command('loadworkflowsampledata', invalid_arg,
                         stderr=invalid_stderr)
            self.assertIn(invalid_message, invalid_stderr.getvalue())

        # Loading valid sample data should succeed without errors
        v1_str = '{}/{}'.format(TEST_WORKFLOW, VERSION_1)
        stdout = StringIO()
        call_command('loadworkflowsampledata', v1_str, stdout=stdout)
        output = stdout.getvalue()
        success_message = 'Successfully loaded sample data'
        self.assertIn(success_message, output)

        # clean up for subsequent commands
        del sys.modules['orchestra.tests.workflows.test_dir.load_sample_data']

        # Loading sample data for a nonexistent workflow should fail
        stderr1 = StringIO()
        call_command(
            'loadworkflowsampledata',
            '{}/{}'.format(NONEXISTENT_WORKFLOW_SLUG, VERSION_1),
            stderr=stderr1)
        output1 = stderr1.getvalue()
        no_workflow_error = ('Workflow {} has not been loaded into the '
                             'database'.format(NONEXISTENT_WORKFLOW_SLUG))
        self.assertIn(no_workflow_error, output1)

        # Loading sample data for a nonexistent version should fail
        stderr2 = StringIO()
        call_command(
            'loadworkflowsampledata',
            '{}/{}'.format(TEST_WORKFLOW, NONEXISTENT_VERSION),
            stderr=stderr2)
        output2 = stderr2.getvalue()
        no_version_error = ('Version {} does not exist'
                            .format(NONEXISTENT_VERSION))
        self.assertIn(no_version_error, output2)

        # Loading a workflow with no loading script should fail
        # Simulate this by moving the file.
        workflow = Workflow.objects.get(slug=TEST_WORKFLOW)
        workflow.sample_data_load_function = 'invalid_load_function'
        workflow.save()
        stderr3 = StringIO()
        call_command('loadworkflowsampledata', v1_str, stderr=stderr3)
        output3 = stderr3.getvalue()
        no_module_error = 'An error occurred while loading sample data'
        self.assertIn(no_module_error, output3)

        # Loading sample data for a workflow with no load function in its JSON
        # manifest should fail.
        workflow = Workflow.objects.get(slug=TEST_WORKFLOW)
        workflow.sample_data_load_function = None
        workflow.save()
        stderr4 = StringIO()
        call_command('loadworkflowsampledata', v1_str, stderr=stderr4)
        output4 = stderr4.getvalue()
        no_load_function_error = ('Workflow {} does not provide sample data'
                                  .format(TEST_WORKFLOW))
        self.assertIn(no_load_function_error, output4)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-04-05 19:27
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0025_iteration'),
    ]

    operations = [
        migrations.CreateModel(
            name='TimeEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date', models.DateField()),
                ('time_worked', models.DurationField()),
                ('description', models.CharField(blank=True, max_length=200, null=True)),
                ('assignment', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='time_entries', to='orchestra.TaskAssignment')),
            ],
        ),
    ]



from annoying.functions import get_object_or_None
from django.conf import settings
from django.core.urlresolvers import reverse
from django.db import transaction
from django.db.models import Q
from django.utils import timezone
from markdown2 import markdown

from orchestra.bots.errors import StaffingResponseException
from orchestra.core.errors import TaskStatusError
from orchestra.core.errors import TaskAssignmentError
from orchestra.bots.staffbot import StaffBot
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import StaffingResponse
from orchestra.models import TaskAssignment
from orchestra.models import Worker
from orchestra.utils.notifications import message_experts_slack_group
from orchestra.utils.task_lifecycle import reassign_assignment
from orchestra.utils.task_lifecycle import assign_task

from orchestra.utils.task_lifecycle import check_worker_allowed_new_assignment
from orchestra.utils.task_lifecycle import get_role_from_counter
from orchestra.utils.task_lifecycle import is_worker_certified_for_task


@transaction.atomic
def handle_staffing_response(worker, staffing_request_inquiry_id,
                             is_available=False):
    """
    Args:
        worker (orchestra.models.Worker):
            Worker instance that responsed to a staffing request inquiry
        staffing_request_inquiry_id (int):
            Id of a staffing_request_inquiry that is associated with a response
        is_available (boolean):
            Boolean that tells whether worker accepted an inquiry or not
    Returns:
        response (orchestra.models.StaffingResponse):
            StaffingResponse object that has been created for the worker
    """

    # TODO(kkamalov): add proper docstring
    staffing_request_inquiry = get_object_or_None(
        StaffingRequestInquiry,
        communication_preference__worker=worker,
        id=staffing_request_inquiry_id
    )
    if staffing_request_inquiry is None:
        return None

    response = (StaffingResponse.objects
                .filter(request_inquiry=staffing_request_inquiry))
    if response.exists():
        response = response.first()
        if not is_available and response.is_winner:
            raise StaffingResponseException(
                'Cannot reject after accepting the task')

        response.is_available = is_available

    else:
        response = StaffingResponse.objects.create(
            request_inquiry=staffing_request_inquiry,
            is_available=is_available)

    if (is_available and
            not StaffingResponse.objects.filter(
                request_inquiry__request=staffing_request_inquiry.request,
                is_winner=True).exists()):
        response.is_winner = True
        request = staffing_request_inquiry.request

        task_assignment = get_object_or_None(
            TaskAssignment,
            task=request.task,
            assignment_counter=request.required_role_counter
        )

        # if task assignment exists then reassign
        if task_assignment is not None:
            reassign_assignment(worker.id, task_assignment.id)
        # otherwise assign task
        else:
            assign_task(worker.id, request.task.id)

    response.save()
    check_responses_complete(staffing_request_inquiry.request)
    return response


def check_responses_complete(request):
    # check all responses have been complete
    responses = StaffingResponse.objects.filter(
        request_inquiry__request=request)
    request_inquiries = StaffingRequestInquiry.objects.filter(
        request=request)
    if (responses.count() == request_inquiries.count() and
            not responses.filter(is_winner=True).exists()):
        # notify that all workers have rejected a task
        message_experts_slack_group(
            request.task.project.slack_group_id,
            ('No worker has accepted to work on task {}'
             .format(request.task)))


def send_staffing_requests(
        worker_batch_size=settings.ORCHESTRA_STAFFBOT_WORKER_BATCH_SIZE,
        frequency=settings.ORCHESTRA_STAFFBOT_BATCH_FREQUENCY):
    staffbot = StaffBot()
    cutoff_datetime = timezone.now() - frequency
    requests = (
        StaffBotRequest.objects
        .filter(status=StaffBotRequest.Status.PROCESSING.value)
        .filter(Q(last_inquiry_sent__isnull=True) |
                Q(last_inquiry_sent__lte=cutoff_datetime)))

    for request in requests:
        send_request_inquiries(staffbot, request, worker_batch_size)


def send_request_inquiries(staffbot, request, worker_batch_size):
    # Get Workers that haven't already received an inquiry.
    workers_with_inquiries = (StaffingRequestInquiry.objects.filter(
        request=request).distinct().values_list(
            'communication_preference__worker__id', flat=True))
    # Sort Workers by their staffing priority first, and then randomly
    # within competing staffing priorities.
    workers = (Worker.objects
               .exclude(id__in=workers_with_inquiries)
               .order_by('-staffing_priority', '?'))
    required_role = get_role_from_counter(request.required_role_counter)
    inquiries_sent = 0

    for worker in workers:
        try:
            check_worker_allowed_new_assignment(worker)
            if (is_worker_certified_for_task(worker, request.task,
                                             required_role,
                                             require_staffbot_enabled=True) and
                    not request.task.is_worker_assigned(worker)):
                staffbot.send_task_to_worker(worker, request)
                inquiries_sent += 1
            if inquiries_sent >= worker_batch_size:
                break

        except TaskStatusError:
            pass
        except TaskAssignmentError:
            pass

    # check whether all inquiries have been sent out.
    if inquiries_sent < worker_batch_size:
        message_experts_slack_group(
            request.task.project.slack_group_id,
            ('All staffing requests for task {} have been sent!'
             .format(request.task)))
        request.status = StaffBotRequest.Status.COMPLETE.value
    request.last_inquiry_sent = timezone.now()
    request.save()


def get_available_requests(worker):
    # We want to show a worker only requests for which there is no
    # winner or for which they have not already replied.
    won_responses = StaffingResponse.objects.filter(is_winner=True)
    worker_provided_responses = StaffingResponse.objects.filter(
        request_inquiry__communication_preference__worker=worker)
    remaining_requests = (
        StaffBotRequest.objects
        .filter(inquiries__communication_preference__worker=worker)
        .exclude(inquiries__responses__in=won_responses)
        .exclude(inquiries__responses__in=worker_provided_responses)
        .distinct())
    inquiries = (
        StaffingRequestInquiry.objects
        .filter(request__in=remaining_requests)
        .filter(communication_preference__worker=worker)
        .order_by('request__task__start_datetime'))
    # Because we might send multiple request inquiries to the same
    # worker for the same request (e.g., email and slack), we
    # deduplicate the inquiries so that we will return at most one
    # inquiry's worth of content here.
    request_ids = set()
    contexts = []
    staffbot = StaffBot()
    for inquiry in inquiries:
        if inquiry.request.id in request_ids:
            continue
        request_ids.add(inquiry.request.id)
        metadata = staffbot.get_staffing_request_metadata(inquiry)
        metadata['detailed_description'] = markdown(
            metadata['detailed_description'])
        metadata['reject_url'] += '?next={}'.format(
            reverse('orchestra:communication:available_staffing_requests'))
        contexts.append(metadata)
    return contexts



from django.test.utils import override_settings
from unittest.mock import patch
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.utils.machine_step_scheduler import (
    AsynchronousMachineStepScheduler)
from orchestra.utils.machine_step_scheduler import MachineStepScheduler


class MachineStepSchedulerTestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()

    def test_machine_step_schedule(self):
        scheduler = MachineStepScheduler()
        with self.assertRaises(NotImplementedError):
            scheduler.schedule(1, '')

    @patch('orchestra.utils.machine_step_scheduler.SynchronousMachineStepScheduler.schedule')  # noqa
    def test_asynchronous_scheduler(self, mock_schedule):
        scheduler = AsynchronousMachineStepScheduler()
        scheduler.schedule(1, 'step1')
        mock_schedule.assert_called_once_with(1, 'step1')

    @patch('orchestra.utils.machine_step_scheduler.schedule_function')
    @override_settings(PRODUCTION=True, WORK_QUEUE=None)
    def test_asynchronous_scheduler_prod(self, mock_schedule):
        scheduler = AsynchronousMachineStepScheduler()
        scheduler.schedule(1, 'step1')
        mock_schedule.assert_called_once_with(None, 'machine_task_executor', 1,
                                              'step1')



from django.views.decorators.csrf import csrf_exempt
from orchestra.project_api.auth import IsSignedUser
from orchestra.project_api.auth import OrchestraProjectAPIAuthentication
from jsonview.decorators import json_view
from rest_framework.decorators import api_view
from rest_framework.decorators import authentication_classes
from rest_framework.decorators import permission_classes

import logging
logger = logging.getLogger(__name__)


def api_exception_logger(func):
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logger.exception('An API exception occurred')
            raise e
    return func_wrapper


def api_endpoint(methods):
    def api_endpoint_decorator(func):
        @csrf_exempt
        @api_view(methods)
        @authentication_classes((OrchestraProjectAPIAuthentication,))
        @permission_classes((IsSignedUser,))
        @json_view
        @api_exception_logger
        def func_wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return func_wrapper
    return api_endpoint_decorator



from orchestra.core.errors import AssignmentPolicyError
from orchestra.core.errors import WorkerCertificationError
from orchestra.models import Task
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_properties import assignment_history

import logging
logger = logging.getLogger(__name__)


def anyone_certified(task, **kwargs):
    """
    Default assignment policy, we leave the task
    in the awaiting processing pool.
    """
    return task


def previously_completed_steps(task, related_steps, **kwargs):
    """
    Assign a new task to the entry-level worker of the specified tasks.
    If no worker can be assigned, return the unmodified task.

    Args:
        task (orchestra.models.Task):
            The newly created task to assign.
        related_steps ([str]):
            List of step slugs from which to attempt to assign a worker.

    Returns:
        task (orchestra.models.Task): The modified task object.

    Raises:
        orchestra.core.errors.AssignmentPolicyError:
            Machine steps cannot be included in an assignment policy.
    """
    if related_steps is None:
        raise AssignmentPolicyError('No related steps given')

    workflow_version = task.step.workflow_version
    for step_slug in related_steps:
        step = workflow_version.steps.get(slug=step_slug)
        if not step.is_human:
            raise AssignmentPolicyError('Machine step should not be '
                                        'member of assignment policy')
    related_tasks = (
        Task.objects
        .filter(step__slug__in=related_steps, project=task.project)
        .select_related('step'))
    for related_task in related_tasks:
        entry_level_assignment = assignment_history(related_task).first()
        if entry_level_assignment and entry_level_assignment.worker:
            try:
                return assign_task(entry_level_assignment.worker.id, task.id)
            except WorkerCertificationError:
                # Task could not be assigned to related worker, try with
                # another related worker
                logger.warning('Tried to assign worker %s to step %s, for '
                               'which they are not certified',
                               entry_level_assignment.worker.id,
                               task.step.slug, exc_info=True)
            except Exception:
                logger.warning('Unable to assign task.', exc_info=True)
    return task



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='project',
            field=models.ForeignKey(default=0, to='orchestra.Project'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='process',
            name='description',
            field=models.TextField(blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='step',
            name='depends_on',
            field=models.ManyToManyField(
                related_name='depends_on_rel_+', to='orchestra.Step', blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='step',
            name='description',
            field=models.TextField(blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='step',
            name='required_certifications',
            field=models.ManyToManyField(
                to='orchestra.Certification', blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='step',
            name='review_policy',
            field=jsonfield.fields.JSONField(blank=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='step',
            name='user_interface',
            field=jsonfield.fields.JSONField(blank=True),
            preserve_default=True,
        ),
    ]



from orchestra.models import Iteration
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.utils.task_properties import current_assignment
from orchestra.utils.task_properties import get_iteration_history


def verify_iterations(task_id):
    task = Task.objects.get(id=task_id)
    iterations = list(get_iteration_history(task).all())
    if iterations:
        _verify_iteration_topology(iterations)
        _verify_iteration_data(iterations)
        _verify_iteration_datetimes(iterations)


def _verify_iteration_topology(iterations):
    # First iteration should belong to first assignment
    expected_counter = 0
    visited_counters = set()
    task = iterations[0].assignment.task
    for i, iteration in enumerate(iterations):
        assignment = iteration.assignment
        assignment_counter = assignment.assignment_counter
        assert assignment_counter == expected_counter
        visited_counters.add(assignment.assignment_counter)

        if i == len(iterations) - 1:
            _verify_final_iteration(iteration)
        else:
            # Only the last iteration (if any) should be processing
            assert iteration.status != Iteration.Status.PROCESSING

        # Status of current iteration determines the expected review level
        # of the next one's assignment
        if iteration.status == Iteration.Status.REQUESTED_REVIEW:
            expected_counter = assignment_counter + 1
        elif iteration.status == Iteration.Status.PROVIDED_REVIEW:
            expected_counter = assignment_counter - 1

    # Iterations should span all assignments
    assert visited_counters == set(range(task.assignments.count()))


def _verify_final_iteration(iteration):
    # Last iteration should belong to current assignment
    assignment = iteration.assignment
    assert assignment == current_assignment(assignment.task)

    # Map final iteration statuses onto task statuses
    task_statuses = {
        Iteration.Status.PROCESSING: [
            Task.Status.PROCESSING, Task.Status.REVIEWING,
            Task.Status.POST_REVIEW_PROCESSING
        ],
        Iteration.Status.REQUESTED_REVIEW: [
            Task.Status.PENDING_REVIEW, Task.Status.COMPLETE
        ],
        Iteration.Status.PROVIDED_REVIEW: [
            Task.Status.POST_REVIEW_PROCESSING
        ]
    }

    # A task awaiting processing should not have iterations
    assignment.task.status != Task.Status.AWAITING_PROCESSING

    for k, v in task_statuses.items():
        # An aborted task could have any iteration configuration
        task_statuses[k].append(Task.Status.ABORTED)

    if iteration.status == Iteration.Status.PROCESSING:
        expected_assignment_status = TaskAssignment.Status.PROCESSING
    else:
        expected_assignment_status = TaskAssignment.Status.SUBMITTED

    # Check that task and assignment statuses are correctly set
    assert assignment.status == expected_assignment_status
    assert assignment.task.status in task_statuses[iteration.status]


def _verify_iteration_data(iterations):
    """
    Verifies correct data for certain iterations.

    Since the data for other iterations won't be stored elsewhere, this
    function should be run each time an iteration is added.
    """
    for iteration in iterations:
        if iteration.status == Iteration.Status.PROCESSING:
            # Iterations should not have data until submitted
            assert iteration.submitted_data == {}
        # NOTE(jrbotros): Last iteration for an assignment will normally
        # have its latest data, unless the task has been reverted


def _verify_iteration_datetimes(iterations):
    """
    Verifies correct start and end datetimes for ordered iterations.
    """
    for iteration in iterations:
        assignment = iteration.assignment
        siblings = assignment.iterations.order_by('start_datetime')
        if siblings.first() == iteration:
            # If iteration is first in assignment, expected start datetime
            # is when the assignment was picked up rather than the end of
            # the previous iteration
            expected_start_datetime = assignment.start_datetime

        assert iteration.start_datetime == expected_start_datetime
        # The expected start datetime for the next iteration should be the
        # end datetime of the current one, unless the next iteration is the
        # first in its assignment
        expected_start_datetime = iteration.end_datetime

        # If iteration is processing, it should not have an end datetime
        if iteration.status == Iteration.Status.PROCESSING:
            assert not iteration.end_datetime



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-08 15:44
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0028_snapshots_to_iterations'),
    ]

    operations = [
        migrations.AddField(
            model_name='timeentry',
            name='worker',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='time_entries', to='orchestra.Worker'),
        ),
        migrations.AlterField(
            model_name='timeentry',
            name='assignment',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='time_entries', to='orchestra.TaskAssignment'),
        ),
    ]



"""
This is the Orchestra Project Python API.

TODO(marcua): Move this file to its own pip/github project.
"""

import json
import requests

from datetime import datetime
from django.conf import settings
from httpsig.requests_auth import HTTPSignatureAuth
from time import mktime
from wsgiref.handlers import format_date_time

import logging
logger = logging.getLogger(__name__)

_httpsig_auth = HTTPSignatureAuth(key_id=settings.ORCHESTRA_PROJECT_API_KEY,
                                  secret=settings.ORCHESTRA_PROJECT_API_SECRET,
                                  algorithm='hmac-sha256')
_api_root_url = '{}/orchestra/api/project/'.format(settings.ORCHESTRA_URL)


class OrchestraError(Exception):
    pass


def _make_api_request(method, endpoint, *args, **kwargs):
    func = getattr(requests, method)
    # Adding 'date' header as per
    # https://github.com/zzsnzmn/py-http-signature/blob/e2e2c753db7da45fab4b215d84e8d490bd708833/http_signature/sign.py#L155  # noqa
    headers = {'date': format_date_time(mktime(datetime.now().timetuple())),
               'X-Api-Version': '~6.5',
               'X-Api-Key': settings.ORCHESTRA_PROJECT_API_KEY}
    headers.update(kwargs.pop('headers', {}))
    all_kwargs = {'auth': _httpsig_auth, 'headers': headers}
    all_kwargs.update(kwargs)
    url = '{}{}/'.format(_api_root_url, endpoint)
    response = func(url, *args, **all_kwargs)
    if response.status_code != 200:
        raise OrchestraError(response.text)
    return response


def get_workflow_types():
    response = _make_api_request('get', 'workflow_types')
    return json.loads(response.text)['workflows']


def get_project_details_url(project_id):
    data = {'project_id': project_id}
    response = _make_api_request('post', 'project_details_url',
                                 data=json.dumps(data))
    url = json.loads(response.text)['project_details_url']
    return url


def create_orchestra_project(client,
                             workflow_slug,
                             workflow_version_slug,
                             description,
                             priority,
                             project_data,
                             task_class):
    data = {
        'workflow_slug': workflow_slug,
        'workflow_version_slug': workflow_version_slug,
        'description': description,
        'priority': priority,
        'project_data': project_data,
        'task_class': task_class
    }
    response = _make_api_request('post', 'create_project',
                                 data=json.dumps(data))
    project_id = json.loads(response.text)['project_id']
    return project_id


def get_project_information(project_id):
    data = {
        'project_id': project_id
    }
    response = _make_api_request('post', 'project_information',
                                 data=json.dumps(data))
    return json.loads(response.text)


def assign_worker_to_task(worker_id, task_id):
    data = {
        'worker_id': worker_id,
        'task_id': task_id,
    }
    response = _make_api_request('post', 'assign_worker_to_task',
                                 data=json.dumps(data))
    return json.loads(response.text)



from datetime import timedelta
from django.conf import settings
from django.test import override_settings
from unittest.mock import patch

from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.tests.helpers.fixtures import StaffBotRequestFactory
from orchestra.tests.helpers.fixtures import StaffingRequestInquiryFactory
from orchestra.tests.helpers.fixtures import StepFactory
from orchestra.tests.helpers.fixtures import TaskFactory
from orchestra.tests.helpers.fixtures import WorkerFactory
from orchestra.bots.errors import SlackUserUnauthorized
from orchestra.bots.staffbot import StaffBot
from orchestra.bots.tests.fixtures import get_mock_slack_data
from orchestra.communication.staffing import send_staffing_requests
from orchestra.communication.mail import html_from_plaintext
from orchestra.models import CommunicationPreference
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import Task
from orchestra.models import Worker
from orchestra.models import WorkerCertification
from orchestra.utils.task_lifecycle import assign_task
from orchestra.utils.task_lifecycle import is_worker_certified_for_task


def _noop_details(task_details):
    return ''


class StaffBotTest(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = self.workers[0]
        self.worker.user.is_superuser = True
        self.worker.user.save()
        patcher = patch(
            ('orchestra.bots.staffbot.StaffBot'
             '._send_staffing_request_by_slack'))
        patcher.start()
        self.addCleanup(patcher.stop)

    def _get_worker_for_task(self, task, role):
        # Get certified reviewer
        for worker in Worker.objects.all():
            if is_worker_certified_for_task(worker, task, role):
                return worker

    def _test_staffing_requests(self, worker, task, command,
                                can_slack=False, can_mail=False):
        StaffBotRequest.objects.all().delete()
        bot = StaffBot()
        communication_type = (CommunicationPreference.CommunicationType
                              .NEW_TASK_AVAILABLE.value)
        communication_preference = CommunicationPreference.objects.get(
            worker=worker,
            communication_type=communication_type)
        communication_preference.methods.slack = can_slack
        communication_preference.methods.email = can_mail
        communication_preference.save()
        data = get_mock_slack_data(
            text=command,
            user_id=self.worker.slack_user_id)
        bot.dispatch(data)
        send_staffing_requests(worker_batch_size=20,
                               frequency=timedelta(minutes=0))
        self.assertEquals(StaffingRequestInquiry.objects.filter(
            communication_preference__worker_id=worker,
            request__task=task).count(), can_slack + can_mail)

    def test_assert_validate_error(self):
        bot = StaffBot()
        with self.assertRaises(SlackUserUnauthorized):
            mock_slack_data = get_mock_slack_data(text='staff 5')
            bot.dispatch(mock_slack_data)

    def test_commands(self):
        """
        Ensure that the bot can handle the following commands:
        /staffbot staff <task_id>
        /staffbot restaff <task_id> <username>

        This test only validates that the commands are processed, other
        tests verify the functionality of the command execution.
        """
        bot = StaffBot()

        # Test staff command
        mock_slack_data = get_mock_slack_data(
            text='staff 5',
            user_id=self.worker.slack_user_id)

        response = bot.dispatch(mock_slack_data)
        self.assertFalse(bot.default_error_text in response.get('text', ''))

        # Test the restaff command
        mock_slack_data['text'] = 'restaff 5 username'
        response = bot.dispatch(mock_slack_data)
        self.assertFalse(bot.default_error_text in response.get('text', ''))

        # Test we fail gracefully
        mock_slack_data['text'] = 'invalid command'
        response = bot.dispatch(mock_slack_data)
        self.assertTrue(bot.default_error_text in response.get('text', ''))

    @patch('orchestra.bots.staffbot.send_mail')
    @patch('orchestra.bots.staffbot.message_experts_slack_group')
    @patch('orchestra.bots.staffbot.StaffBot._send_staffing_request_by_slack')
    def test_staff_command(self, mock_slack, mock_experts_slack, mock_mail):
        """
        Test that the staffing logic is properly executed for the
        staff command.
        """
        task = (Task.objects
                .filter(status=Task.Status.AWAITING_PROCESSING)
                .first())

        # Get certified worker
        worker = self._get_worker_for_task(
            task, WorkerCertification.Role.ENTRY_LEVEL)

        self._test_staffing_requests(worker, task, 'staff {}'.format(task.id),
                                     can_slack=True, can_mail=True)

        self._test_staffing_requests(worker, task, 'staff {}'.format(task.id),
                                     can_slack=False, can_mail=False)

        # Change the task state to pending review
        task = assign_task(worker.id, task.id)
        task.status = Task.Status.PENDING_REVIEW
        task.save()

        StaffingRequestInquiry.objects.all().delete()

        worker = self._get_worker_for_task(task,
                                           WorkerCertification.Role.REVIEWER)
        self._test_staffing_requests(worker, task, 'staff {}'.format(task.id),
                                     can_slack=False, can_mail=False)
        self._test_staffing_requests(worker, task, 'staff {}'.format(task.id),
                                     can_slack=True, can_mail=True)
        self.assertTrue(mock_mail.called)
        self.assertTrue(mock_experts_slack.called)
        self.assertTrue(mock_slack.called)

    def test_staff_command_errors(self):
        """
        Test that the staffing logic errors are raised during
        staff command.
        """
        bot = StaffBot()
        data = get_mock_slack_data(
            text='staff 999999999999',
            user_id=self.worker.slack_user_id)

        response = bot.dispatch(data)
        self.assertEqual(response['attachments'][0]['text'],
                         bot.task_does_not_exist_error.format('999999999999'))

        data['text'] = 'staff'
        response = bot.dispatch(data)
        self.assertTrue(bot.default_error_text in response.get('text'))

        task = TaskFactory(status=Task.Status.COMPLETE)
        data['text'] = 'staff {}'.format(task.id)
        response = bot.dispatch(data)
        self.assertEquals(response['attachments'][0]['text'],
                          bot.task_assignment_error
                          .format(task.id,
                                  'Status incompatible with new assignment'))

    @patch('orchestra.bots.staffbot.send_mail')
    @patch('orchestra.bots.staffbot.message_experts_slack_group')
    @patch('orchestra.bots.staffbot.StaffBot._send_staffing_request_by_slack')
    def test_restaff_command(self, mock_slack, mock_experts_slack, mock_mail):
        """
        Test that the restaffing logic is properly executed for the
        restaff command.
        """
        task = (Task.objects
                .filter(status=Task.Status.AWAITING_PROCESSING)
                .first())

        # Get certified worker
        task = assign_task(self.worker.id, task.id)
        command = 'restaff {} {}'.format(task.id, self.worker.user.username)

        worker = self.workers[3]
        self._test_staffing_requests(worker, task, command,
                                     can_slack=False, can_mail=True)
        self.assertTrue(mock_mail.called)
        self.assertTrue(mock_experts_slack.called)
        self.assertTrue(mock_slack.called)

    def test_restaff_command_errors(self):
        """
        Test that the staffing logic errors are raised during
        staff command.
        """
        bot = StaffBot()
        command = 'restaff 999999999999 unknown'
        data = get_mock_slack_data(
            text=command,
            user_id=self.worker.slack_user_id)

        response = bot.dispatch(data)
        self.assertEqual(response.get('text'),
                         command)

        self.assertEqual(response['attachments'][0]['text'],
                         bot.worker_does_not_exist.format('unknown'))

        worker = WorkerFactory(user__username='username')
        data['text'] = 'restaff 999999999999 username'
        response = bot.dispatch(data)
        self.assertEqual(response['attachments'][0]['text'],
                         bot.task_does_not_exist_error.format('999999999999'))

        # making sure it works with slack username as well.
        worker.slack_username = 'slackusername'
        worker.save()
        data['text'] = 'restaff 999999999999 slackusername'
        response = bot.dispatch(data)
        self.assertEqual(response['attachments'][0]['text'],
                         bot.task_does_not_exist_error.format('999999999999'))

        data['text'] = 'restaff'
        response = bot.dispatch(data)
        self.assertTrue(bot.default_error_text in response.get('text'))

        task = TaskFactory(status=Task.Status.COMPLETE)
        command = 'restaff {} {}'.format(task.id, worker.user.username)

        data['text'] = command
        response = bot.dispatch(data)
        self.assertEquals(response['attachments'][0]['text'],
                          (bot.task_assignment_does_not_exist_error
                           .format(worker.user.username, task.id)))

    @override_settings(ORCHESTRA_MOCK_EMAILS=True)
    @patch('orchestra.bots.staffbot.send_mail')
    def test_get_staffing_request_messsage(self, mock_mail):
        def _task_factory(status, path):
            description_no_kwargs = {'path': path}
            return TaskFactory(
                status=status,
                step=StepFactory(
                    slug='stepslug',
                    description='the step',
                    detailed_description_function=description_no_kwargs),
                project__workflow_version__workflow__description=(
                    'the workflow'),
                project__short_description='the coolest project'
            )

        # Test slack without review and with a detailed_description_function
        task = _task_factory(
            Task.Status.AWAITING_PROCESSING,
            'orchestra.tests.helpers.fixtures.get_detailed_description')
        staffing_request_inquiry = StaffingRequestInquiryFactory(
            communication_preference__worker__user__first_name='test-name',
            request__task=task)
        message = StaffBot()._get_staffing_request_message(
            staffing_request_inquiry,
            'communication/new_task_available_slack.txt')
        self.assertEqual(message,
                         '''Hello test-name!

A new task is available for you to work on, if you'd like!  Here are the details:

Project: the workflow
Project description: the coolest project
Task: the step
Details: No text given stepslug

<http://127.0.0.1:8000/orchestra/communication/accept_staffing_request_inquiry/{}/|Accept the Task>
<http://127.0.0.1:8000/orchestra/communication/reject_staffing_request_inquiry/{}/|Ignore the Task>
<http://127.0.0.1:8000/orchestra/communication/available_staffing_requests/|View All Available Tasks>

'''.format(staffing_request_inquiry.id, staffing_request_inquiry.id))  # noqa

        # Test email with review and no detailed_description_function
        task = _task_factory(
            Task.Status.PENDING_REVIEW,
            'orchestra.bots.tests.test_staffbot._noop_details')
        staffing_request_inquiry = StaffingRequestInquiryFactory(
            communication_preference__worker__user__first_name='test-name2',
            request=StaffBotRequestFactory(
                task=task, required_role_counter=1))
        message = StaffBot()._get_staffing_request_message(
            staffing_request_inquiry,
            'communication/new_task_available_email.txt')
        self.assertEqual(message,
                         '''Hello test-name2!

A new task is available for you to work on, if you'd like!  Here are the details:

Project: the workflow
Project description: the coolest project
Task: the step [Review]


<a href="http://127.0.0.1:8000/orchestra/communication/accept_staffing_request_inquiry/{}/">Accept the Task</a>
<a href="http://127.0.0.1:8000/orchestra/communication/reject_staffing_request_inquiry/{}/">Ignore the Task</a>
<a href="http://127.0.0.1:8000/orchestra/communication/available_staffing_requests/">View All Available Tasks</a>

'''.format(staffing_request_inquiry.id, staffing_request_inquiry.id))  # noqa

        # Test that we markdown things
        StaffBot()._send_staffing_request_by_mail('test@test.com', message)
        mock_mail.assert_called_once_with(
            'A new task is available for you',
            message,
            settings.ORCHESTRA_NOTIFICATIONS_FROM_EMAIL,
            ['test@test.com'],
            html_message=html_from_plaintext(message)
        )



class AssignmentPolicyError(Exception):
    pass


class IllegalTaskSubmission(Exception):
    pass


class InvalidSlugValue(Exception):
    pass


class MachineExecutionError(Exception):
    pass


class ModelSaveError(Exception):
    pass


class NoTaskAvailable(Exception):
    pass


class ReviewPolicyError(Exception):
    pass


class S3UploadError(Exception):
    pass


class SlugUniquenessError(Exception):
    pass


class InvalidRevertError(Exception):
    pass


class TaskAssignmentError(Exception):
    pass


class TaskDependencyError(Exception):
    pass


class TaskStatusError(Exception):
    pass


class WorkerCertificationError(Exception):
    pass


class WorkflowError(Exception):
    pass


class TimerError(Exception):
    pass



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-26 20:34
from __future__ import unicode_literals

import bitfield.models
from django.db import migrations, models
import django.db.models.deletion
import orchestra.models.communication.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0044_auto_20160426_0044'),
    ]

    operations = [
        migrations.CreateModel(
            name='CommunicationPreference',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('methods', bitfield.models.BitField((('slack', 'Slack'), ('email', 'Email')), blank=True, default=None, null=True)),
                ('communication_type', models.IntegerField(choices=[(0, 'task_status_change')])),
                ('worker', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='orchestra.Worker')),
            ],
            bases=(orchestra.models.communication.mixins.CommunicationPreferenceMixin, models.Model),
        ),
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='communicationpreference',
            unique_together=set([('worker', 'communication_type')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-26 20:34
from __future__ import unicode_literals

from enum import Enum

from django.db import migrations

# Copy pasted from models since migrations does not have access
class CommunicationMethods(object):
    SLACK = 'slack'
    EMAIL = 'email'


COMMUNICATION_METHODS = (
    (CommunicationMethods.SLACK, 'Slack'),
    (CommunicationMethods.EMAIL, 'Email'),
)

class CommunicationType(Enum):
    TASK_STATUS_CHANGE = 0

def add_communcation_type_to_workers(apps, schema_editor):
    CommunicationPreference = apps.get_model(
        'orchestra', 'CommunicationPreference')
    Worker = apps.get_model('orchestra', 'Worker')

    default_methods = 2**len(COMMUNICATION_METHODS) - 1
    for worker in Worker.objects.all():
        CommunicationPreference.objects.get_or_create(
            worker=worker,
            methods=default_methods,
            communication_type=CommunicationType.TASK_STATUS_CHANGE.value)


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0045_auto_20160426_2034'),
    ]

    operations = [
        migrations.RunPython(add_communcation_type_to_workers)  # manually-reviewed
    ]



from django.db import models


class CommunicationPreferenceManager(models.Manager):

    def get_or_create_all_types(self, worker, methods=None):
        """
            Create CommunicationPreference objects with all of the given
            CommunicationTypes and the given methods flags.
        """

        methods = methods or self.model.get_default_methods()
        choices = self.model.CommunicationType.choices()
        comm_prefs = []
        for communication_type, _ in choices:
            comm_pref, _ = self.get_or_create(
                worker=worker,
                methods=methods,
                communication_type=communication_type
            )
            comm_prefs.append(comm_prefs)
        return comm_prefs



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-05-15 15:45
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0061_auto_20160505_2126'),
    ]

    operations = [
        migrations.RenameField(
            model_name='staffingresponse',
            old_name='request',
            new_name='request_inquiry',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-26 00:44
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0043_tasktimer_description'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='timeentry',
            options={'verbose_name_plural': 'time entries'},
        ),
    ]



from pydoc import locate
from django.conf import settings

from beanstalk_dispatch import ARGS
from beanstalk_dispatch import FUNCTION
from beanstalk_dispatch import KWARGS
from beanstalk_dispatch import BeanstalkDispatchError


def execute_function(function_request):
    """
    Given a request created by
    `beanstalk_dispatch.common.create_request_body`, executes the
    request.  This function is to be run on a beanstalk worker.
    """
    dispatch_table = getattr(settings, 'BEANSTALK_DISPATCH_TABLE', None)

    if dispatch_table is None:
        raise BeanstalkDispatchError('No beanstalk dispatch table configured')
    for key in (FUNCTION, ARGS, KWARGS):
        if key not in function_request.keys():
            raise BeanstalkDispatchError(
                'Please provide a {} argument'.format(key))

    function_path = dispatch_table.get(
        function_request[FUNCTION], ''
    )

    if function_path:
        # TODO(marcua): Catch import errors and rethrow them as
        # BeanstalkDispatchErrors.
        function = locate(function_path)
        function(*function_request[ARGS], **function_request[KWARGS])
    else:
        raise BeanstalkDispatchError(
            'Requested function not found: {}'.format(
                function_request[FUNCTION]))



from orchestra.machine_tasks import execute
from beanstalk_dispatch.client import schedule_function
from django.conf import settings


class MachineStepScheduler(object):
    def schedule(self, project_id, step_slug):
        raise NotImplementedError()


class SynchronousMachineStepScheduler(MachineStepScheduler):
    def schedule(self, project_id, step_slug):
        execute(project_id, step_slug)


class AsynchronousMachineStepScheduler(MachineStepScheduler):
    def schedule(self, project_id, step_slug):
        if not settings.PRODUCTION and not settings.STAGING:
            scheduler = SynchronousMachineStepScheduler()
            scheduler.schedule(project_id, step_slug)
        else:
            schedule_function(settings.WORK_QUEUE, 'machine_task_executor',
                              project_id, step_slug)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):
    dependencies = [
        ('orchestra', '0005_auto_20150402_0538'),
    ]

    operations = [
        migrations.RenameField(
            model_name='task',
            old_name='step',
            new_name='step_slug'
        ),
        migrations.RenameField(
            model_name='project',
            old_name='process',
            new_name='process_slug'
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0004_taskassignment_task'),
    ]

    operations = [
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='step',
            unique_together=None,
        ),
        migrations.RemoveField(  # manually-reviewed
            model_name='step',
            name='depends_on',
        ),
        migrations.RemoveField(  # manually-reviewed
            model_name='step',
            name='process',
        ),
        migrations.RemoveField(  # manually-reviewed
            model_name='step',
            name='required_certifications',
        ),
        migrations.AlterField(
            model_name='project',
            name='process',
            field=models.CharField(
                max_length=200, choices=[('Doctors', 'Doctors Process')]),
            preserve_default=True,
        ),
        migrations.DeleteModel(  # manually-reviewed
            name='Process',
        ),
        migrations.AlterField(
            model_name='task',
            name='project',
            field=models.ForeignKey(
                related_name='tasks', to='orchestra.Project'),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='task',
            name='step',
            field=models.CharField(max_length=200, choices=[
                                   ('Content Extraction', ' Content Extraction'), ('Copy Pass', ' Copy Pass')]),
            preserve_default=True,
        ),
        migrations.DeleteModel(  # manually-reviewed
            name='Step',
        ),
        migrations.AlterField(
            model_name='taskassignment',
            name='task',
            field=models.ForeignKey(
                related_name='assignments', to='orchestra.Task'),
            preserve_default=True,
        ),
    ]



from orchestra.google_apps.convenience import create_project_google_folder
from orchestra.models import Project
from orchestra.models import WorkflowVersion
from orchestra.communication.slack import create_project_slack_group
from orchestra.utils.task_lifecycle import create_subsequent_tasks


def create_project_with_tasks(workflow_slug,
                              workflow_version_slug,
                              description,
                              priority,
                              task_class,
                              project_data):

    workflow_version = WorkflowVersion.objects.get(
        slug=workflow_version_slug,
        workflow__slug=workflow_slug)

    project = Project.objects.create(workflow_version=workflow_version,
                                     short_description=description,
                                     priority=priority,
                                     project_data=project_data,
                                     task_class=task_class)

    create_project_slack_group(project)
    create_project_google_folder(project)

    create_subsequent_tasks(project)
    return project



from orchestra.models import CommunicationPreference
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.tests.helpers.fixtures import WorkerFactory


class ManagersTestCase(OrchestraTestCase):

    def setUp(self):
        self.user = UserFactory()
        self.worker = WorkerFactory(user=self.user)
        super().setUp()

    def test_get_or_create_all_types(self):
        comm_prefs = CommunicationPreference.objects.get_or_create_all_types(
            self.worker)
        comm_types = CommunicationPreference.CommunicationType.choices()
        self.assertEqual(len(comm_prefs), len(comm_types))
        for comm_type, _ in comm_types:
            comm_pref = CommunicationPreference.objects.get(
                communication_type=comm_type,
                worker=self.worker)
            for key, is_set in comm_pref.methods.iteritems():
                self.assertTrue(is_set)

    def test_get_or_create_all_types_set_methods(self):
        """
            Manually set the desired methods that are allowed.
        """

        methods = ~CommunicationPreference.methods.email
        methods &= CommunicationPreference.methods.slack
        comm_prefs = CommunicationPreference.objects.get_or_create_all_types(
            self.worker, methods=methods)
        comm_types = CommunicationPreference.CommunicationType.choices()
        self.assertEqual(len(comm_prefs), len(comm_types))
        for comm_type, _ in comm_types:
            comm_pref = CommunicationPreference.objects.get(
                communication_type=comm_type,
                worker=self.worker)
            self.assertTrue(comm_pref.methods.slack)
            self.assertFalse(comm_pref.methods.email)



"""Increments the Orchestra version and pushes a new release to PyPI.

This script can only be called by orchestra core committers, and will result
in permission errors otherwise.
"""
from argparse import ArgumentParser
from distutils.version import StrictVersion
import os
import re
import shutil
from subprocess import check_output
from tempfile import mkdtemp

VERSION_RE = re.compile("__version__ = ['\"]([^'\"]+)['\"]")


def main(args):
    # Compute the new version number
    old_version = get_version('orchestra')
    new_version = increment_version_number(old_version, args.version_part)

    # Verify that our git setup is reasonable and print a loud warning.
    verify_and_warn(old_version, new_version, args)

    # Update orchestra/__init__.py with the new version
    if args.no_commit:
        print('--no-commit passed, not committing new version to Orchestra.')
        release_version = old_version
        version_text = 'current'
    else:
        update_version('orchestra', new_version, fake=args.fake)
        release_version = new_version
        version_text = 'new'

        # Commit the update to the repo
        commit_and_push(fake=args.fake)

    # Create a new tag for the release
    if args.no_tag:
        print('--no-tag passed, not tagging the {} version.'.format(
            version_text))
        tag_str = 'v{}'.format(release_version)
    else:
        tag_str = tag_release(release_version, fake=args.fake)

    # Release the new version on pypi
    if args.no_pypi:
        print('--no-pypi passed, not releasing the {} version to PyPI.'.format(
            version_text))
    else:
        pypi_release(tag_str, fake=args.fake)

    if not args.no_tag:
        print('Congratulations! Version {} has been tagged. You will want to '
              'make sure that the documentation is up to date, by activating '
              'v{} and by forcing a rebuild of the stable version. '
              .format(release_version, release_version))


def verify_and_warn(old_version, new_version, args):
    # verify that we're on the master branch
    branch_cmd = ['git', 'symbolic-ref', '--short', 'HEAD']
    cur_branch = run_command(branch_cmd).strip().lower()
    if cur_branch != 'master':
        print('This script must be run from the master branch. Exiting.')
        exit()

    # verify that master is up to date
    rev_cmd = ['git', 'rev-list']
    unpushed_cmd = rev_cmd + ['origin/master..']
    unpushed_commits = run_command(unpushed_cmd).strip()
    if unpushed_commits:
        print('Your branch has commits not yet on origin/master. Exiting.')
        exit()

    run_command(['git', 'fetch', 'origin', 'master'])
    unpulled_cmd = rev_cmd + ['..origin/master']
    unpulled_commits = run_command(unpulled_cmd).strip()
    if unpulled_commits:
        print('Your branch is behind origin/master. Exiting.')
        exit()

    # verify that there are no outstanding changes
    status_cmd = ['git', 'status', '--porcelain']
    status_output = run_command(status_cmd).strip()
    if status_output:
        print('There are outstanding changes to your branch. Exiting.')
        exit()

    action_text = get_action_text(old_version, new_version, args)
    if not action_text:  # Nothing to warn about.
        return
    print('WARNING: running this script will {}. This involves making changes '
          'you CANNOT TAKE BACK. Before proceeding, ensure that you are on '
          'the master branch, have pulled the latest changes, and have no '
          'outstanding local changes. THIS SCRIPT WILL FAIL if you do not '
          'have credentials to push to the Orchestra GitHub repository or the '
          'Orchestra PyPI account.'.format(action_text))
    ack = input('ARE YOU SURE YOU WISH TO PROCEED? (Type "release" to '
                'confirm): ').lower().strip('\' "')
    while ack in ['y', 'yes']:
        ack = input('Please type "release" to confirm: ')
    if ack != 'release':
        exit()


def get_action_text(old_version, new_version, args):
    actions = []
    release_version = old_version if args.no_commit else new_version
    if not args.no_commit:
        actions.append('increment the Orchestra version from {} to {}'.format(
            old_version, new_version))
    if not args.no_tag:
        actions.append('create a tag for version {}'.format(release_version))
    if not args.no_pypi:
        actions.append('release a PyPI distribution for version {}'.format(
            release_version))
    if not actions:
        return ''

    if len(actions) == 1:
        action_text = actions[0]
    elif len(actions) == 2:
        action_text = ' and '.join(actions)
    else:
        actions[-1] = 'and ' + actions[-1]
        action_text = ', '.join(actions)
    return action_text


def increment_version_number(old_version, version_part):
    version_parsed = StrictVersion(old_version)
    major, minor, patch = version_parsed.version
    if version_part == 'major':
        major += 1
        minor = 0
        patch = 0
    elif version_part == 'minor':
        minor += 1
        patch = 0
    elif version_part == 'patch':
        patch += 1
    new_version = '{}.{}.{}'.format(major, minor, patch)
    return new_version


def version_file(package):
    return os.path.join(package, '__init__.py')


def get_version(package):
    """ Return package version as listed in `__version__` in `init.py`."""
    # Parsing the file instead of importing it Pythonically allows us to make
    # this script completely Django-independent. This function is also used by
    # setup.py, which cannot import the package it is installing.
    with open(version_file(package), 'r') as f:
        init_py = f.read()
    return VERSION_RE.search(init_py).group(1)


def update_version(package, new_version, fake=False):
    # Update the __init__.py file with the new version.
    init_file = version_file(package)
    with open(init_file, 'r') as f:
        old_initpy = f.read()
    new_initpy = VERSION_RE.sub("__version__ = '{}'".format(new_version),
                                old_initpy)

    # Write the file out to disk.
    if fake:
        print('--fake passed, not updating {}'.format(init_file))
    else:
        print('Updating version in {}'.format(init_file))
        with open(version_file(package), 'w') as f:
            f.write(new_initpy)


def commit_and_push(fake=False):
    # commit all changes
    wrap_command(['git', 'commit', '-am', 'Version bump.'], fake)

    # Push changes
    wrap_command(['git', 'push', 'origin', 'master'], fake)


def tag_release(version, fake=False):
    # Create the tag
    tag_str = 'v{}'.format(version)
    tag_msg = 'Version {} of Orchestra'.format(version)
    tag_cmd = ['git', 'tag', '-am', tag_msg, tag_str]
    wrap_command(tag_cmd, fake)

    # Update the stable tag
    stable_tag_msg = 'The latest stable release of Orchestra.'
    stable_tag_cmd = ['git', 'tag', '-afm', stable_tag_msg, 'stable']
    wrap_command(stable_tag_cmd, fake)

    # Push the tags
    wrap_command(['git', 'push', '-f', '--tags'], fake)

    return tag_str


def pypi_release(tag_str, fake=False):
    # Create a temporary directory for the release
    release_dir = mkdtemp()
    print('Created release directory in {}'.format(release_dir))

    # Move into the release directory
    old_dir = os.getcwd()
    os.chdir(release_dir)

    # Clone the git repo for release
    print('Cloning Orchestra into the new release directory.')
    clone_cmd = ['git', 'clone', 'https://github.com/b12io/orchestra',
                 '-b', tag_str]
    wrap_command(clone_cmd, fake)  # This is safe to always run.
    if not fake:
        os.chdir('orchestra')

    # Release to pypi
    print('Releasing Orchestra to PyPI.')
    pypi_cmd = ['python3', 'setup.py', 'sdist', 'upload', '-r', 'pypi']
    wrap_command(pypi_cmd, fake)

    # Clean up
    print('Cleaning up release directory.')
    os.chdir(old_dir)
    shutil.rmtree(release_dir)


def wrap_command(cmd, fake):
    cmd_str = ' '.join(cmd)
    if fake:
        print('--fake passed, not running "{}"'.format(cmd_str))
        return ''
    print('Running command "{}"'.format(cmd_str))
    run_command(cmd)


def run_command(cmd):
    return check_output(cmd).decode()


def parse_args():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        'version_part',
        choices=['major', 'minor', 'patch'],
        help=('The part of the version to increase: major (e.g., 1.0.0 => '
              '2.0.0), minor (e.g., 0.1.0 => 0.2.0), or patch (e.g., 0.0.1 => '
              '0.0.2).'))
    parser.add_argument(
        '--fake',
        action='store_true',
        help=('Fake the release without actually changing anything (useful '
              'for testing).'))
    parser.add_argument(
        '--no-commit',
        action='store_true',
        help=("Don\'t push a commit that updates the version number in the "
              "codebase (i.e., release the current version. Useful for "
              " continuing if the script ran partially before)."))
    parser.add_argument(
        '--no-tag',
        action='store_true',
        help=("Don\'t create a tag for the new version (i.e., release the "
              "current version's tag. Useful for continuing if the script ran "
              "partially before)."))
    parser.add_argument(
        '--no-pypi',
        action='store_true',
        help=("Don\'t create a new pypi release."))

    return parser.parse_args()


if __name__ == '__main__':
    main(parse_args())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0021_merge'),
    ]

    operations = [
        migrations.RenameField(
            model_name='project',
            old_name='review_document_url',
            new_name='team_messages_url',
        ),
        migrations.AlterField(
            model_name='project',
            name='project_data',
            field=jsonfield.fields.JSONField(blank=True, default={}),
        ),
        migrations.AlterField(
            model_name='taskassignment',
            name='in_progress_task_data',
            field=jsonfield.fields.JSONField(blank=True, default={}),
        ),
        migrations.AlterField(
            model_name='taskassignment',
            name='snapshots',
            field=jsonfield.fields.JSONField(blank=True, default={}),
        ),
    ]
